"""
Testing framework for OWWidgets
"""
import os
import sys
import pickle
import tempfile

import time
import unittest
from contextlib import contextmanager, ExitStack
from unittest.mock import Mock, patch
from typing import List, Optional, TypeVar, Type

from AnyQt.QtCore import Qt, QObject, pyqtSignal, QElapsedTimer, pyqtSlot
from AnyQt.QtTest import QTest, QSignalSpy
from AnyQt.QtWidgets import (
    QApplication, QComboBox, QSpinBox, QDoubleSpinBox, QSlider
)
from AnyQt import sip

from orangewidget.report.owreport import OWReport
from orangewidget.settings import SettingsHandler
from orangewidget.utils.signals import get_input_meta, notify_input_helper, \
    Output, Input, LazyValue

from orangewidget.widget import OWBaseWidget

if hasattr(sip, "setdestroyonexit"):
    sip.setdestroyonexit(False)

app = None

DEFAULT_TIMEOUT = 5000

# pylint: disable=invalid-name
T = TypeVar("T")


@contextmanager
def named_file(content, encoding=None, suffix=''):
    file = tempfile.NamedTemporaryFile("wt", delete=False,
                                       encoding=encoding, suffix=suffix)
    file.write(content)
    name = file.name
    file.close()
    try:
        yield name
    finally:
        os.remove(name)


class _Invalidated(QObject):
    completed = pyqtSignal(object)


class _FinishedMonitor(QObject):
    finished = pyqtSignal()

    def __init__(self, widget: 'OWBaseWidget') -> None:
        super().__init__(None)
        self.widget = widget
        self.sm = widget.signalManager  # type: DummySignalManager
        widget.widgetStateChanged.connect(self._changed)
        self._finished = self.is_finished()
        self.invalidated_outputs = self.sm.invalidated_outputs(widget)
        for output in self.invalidated_outputs:
            output.completed.connect(self._completed, Qt.UniqueConnection)

    def is_finished(self) -> bool:
        finished = not (self.widget.isInvalidated() or
                        self.sm.has_invalidated_outputs(self.widget))
        return finished

    @pyqtSlot()
    def _changed(self):
        fin = self.is_finished()
        self.invalidated_outputs = self.sm.invalidated_outputs(self.widget)
        for output in self.invalidated_outputs:
            try:
                output.completed.connect(self._completed, Qt.UniqueConnection)
            except TypeError:  # connection already exists
                pass
        if fin and fin != self._finished:
            self.finished.emit()

    @pyqtSlot()
    def _completed(self):
        self._changed()


class DummySignalManager:
    def __init__(self):
        self.outputs = {}

    def clear(self):
        self.outputs.clear()

    def send(self, widget, signal_name, value, *args, **kwargs):
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        current = self.outputs.get((widget, signal_name), None)
        self.outputs[(widget, signal_name)] = value
        if isinstance(current, _Invalidated):
            current.completed.emit(value)

    def invalidate(self, widget, signal_name):
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        self.outputs[(widget, signal_name)] = _Invalidated()

    def wait_for_outputs(self, widget, timeout=DEFAULT_TIMEOUT):
        st = _Invalidated()
        invalidated = self.invalidated_outputs(widget)
        for val in invalidated:
            val.completed.connect(st.completed)
        if invalidated:
            return QSignalSpy(st.completed).wait(timeout)
        else:
            return True

    def has_invalidated_outputs(self, widget):
        invalidated = self.invalidated_outputs(widget)
        return bool(invalidated)

    def invalidated_outputs(self, widget):
        return [value for (w, name), value in self.outputs.items()
                if w is widget and isinstance(value, _Invalidated)]

    def get_output(self, widget, signal_name, timeout=DEFAULT_TIMEOUT):
        if not isinstance(signal_name, str):
            signal_name = signal_name.name
        elapsed = QElapsedTimer()
        if widget.isInvalidated():
            elapsed.start()
            spy = QSignalSpy(widget.invalidatedStateChanged)
            assert spy.wait(timeout)
            timeout = timeout - elapsed.elapsed()
        value = self.outputs.get((widget, signal_name))
        if isinstance(value, _Invalidated) and timeout >= 0:
            spy = QSignalSpy(value.completed)
            assert spy.wait(timeout), "Failed to get output in the specified timeout"
            assert len(spy) == 1
            value = spy[0][0]
        return value

    def wait_for_finished(
            self, widget: 'OWBaseWidget', timeout=DEFAULT_TIMEOUT) -> bool:
        monitor = _FinishedMonitor(widget)
        if monitor.is_finished():
            return True
        else:
            spy = QSignalSpy(monitor.finished)
            return spy.wait(timeout)


class GuiTest(unittest.TestCase):
    """Base class for tests that require a QApplication instance

    GuiTest ensures that a QApplication exists before tests are run an
    """
    tear_down_stack: ExitStack
    LANGUAGE = "English"

    @classmethod
    def setUpClass(cls):
        """Prepare for test execution.

        Ensure that a (single copy of) QApplication has been created
        """
        global app
        if app is None:
            app = QApplication.instance()
        if app is None:
            app = QApplication(["-", "-widgetcount"])

        # Disable App Nap on macOS (see
        # https://codereview.qt-project.org/c/qt/qtbase/+/202515 for more)
        if sys.platform == "darwin":
            try:
                import appnope
            except ImportError:
                pass
            else:
                appnope.nope()
        cls.tear_down_stack = ExitStack()
        if "pyqtgraph" in sys.modules:
            # undo pyqtgraph excepthook override, abort on exceptions in
            # slots, event handlers, ...
            sys.excepthook = sys.__excepthook__
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        if "pyqtgraph" in sys.modules:
            import pyqtgraph
            pyqtgraph.setConfigOption("exitCleanup", False)
        cls.tear_down_stack.close()
        super().tearDownClass()
        QTest.qWait(0)

    def tearDown(self) -> None:
        """
        Process any pending events before the next test is executed. This
        includes deletes scheduled with `QObject.deleteLater`.
        """
        super().tearDown()
        QTest.qWait(0)

    @classmethod
    def skipNonEnglish(cls, f):
        return cls.runOnLanguage("English")(f)

    @classmethod
    def runOnLanguage(cls, lang):
        def decorator(f):
            if cls.LANGUAGE != lang:
                f = unittest.skip(f"Test is valid only for {lang} release")(f)
            return f
        return decorator


NO_VALUE = object()


class WidgetTest(GuiTest):
    """Base class for widget tests

    Contains helper methods widget creation and working with signals.

    All widgets should be created by the create_widget method, as this
    will ensure they are created correctly.
    """

    widgets = []  # type: List[OWBaseWidget]

    def __init_subclass__(cls, **kwargs):
        def test_minimum_size(self):
            widget = getattr(self, "widget", None)
            if widget is None:
                self.skipTest("minimum size not tested as .widget was not set")
            self.check_minimum_size(widget)

        def test_image_export(self):
            widget = getattr(self, "widget", None)
            if widget is None:
                self.skipTest("image exporting not tested as .widget was not set")
            self.check_export_image(widget)

        def test_msg_base_class(self):
            widget = getattr(self, "widget", None)
            if widget is None:
                self.skipTest("msg base class not tested as .widget was not set")
            self.check_msg_base_class(widget)

        if not hasattr(cls, "test_minimum_size"):
            cls.test_minimum_size = test_minimum_size
        if not hasattr(cls, "test_msg_base_class"):
            cls.test_msg_base_class = test_msg_base_class
        if not hasattr(cls, "test_image_export"):
            cls.test_image_export = test_image_export

    @classmethod
    def setUpClass(cls):
        """Prepare environment for test execution

        Construct a dummy signal manager and monkey patch
        OWReport.get_instance to return a manually created instance.
        """
        super().setUpClass()

        cls.widgets = []

        cls.signal_manager = DummySignalManager()

        report = None

        def get_instance():
            nonlocal report
            if report is None:
                report = OWReport()
                report.have_report_warning_shown = True  # if missing QtWebView/QtWebKit
                if not (os.environ.get("TRAVIS") or os.environ.get("APPVEYOR")):
                    report.show = Mock()
                cls.widgets.append(report)
            return report

        cls.tear_down_stack.enter_context(
            patch.object(OWReport, "get_instance", get_instance)
        )

    @classmethod
    def tearDownClass(cls) -> None:
        cls.signal_manager.clear()
        del cls.signal_manager
        widgets = cls.widgets[:]
        cls.widgets.clear()
        while widgets:
            w = widgets.pop(-1)
            if not w.__dict__.get("_Cls__didCallOnDeleteWidget", False):
                w.onDeleteWidget()
            if not sip.isdeleted(w):
                w.deleteLater()
            w.signalManager = None
        super().tearDownClass()

    def tearDown(self):
        """Process any pending events before the next test is executed."""
        self.signal_manager.clear()
        super().tearDown()

    def create_widget(self, cls, stored_settings=None, reset_default_settings=True):
        # type: (Type[T], Optional[dict], bool) -> T
        """Create a widget instance using mock signal_manager.

        When used with default parameters, it also overrides settings stored
        on disk with default defined in class.

        After widget is created, QApplication.process_events is called to
        allow any singleShot timers defined in __init__ to execute.

        Parameters
        ----------
        cls : WidgetMetaClass
            Widget class to instantiate
        stored_settings : dict
            Default values for settings
        reset_default_settings : bool
            If set, widget will start with default values for settings,
            if not, values accumulated through the session will be used

        Returns
        -------
        Widget instance : cls
        """
        # Use a substitute subclass to mark calls to onDeleteWidget; Some tests
        # call this on their own (this used to be done in tearDownClass, then
        # it was not, so tests did it themself, now it is done again).
        with open_widget_classes():
            class Cls(cls):
                def onDeleteWidget(self):
                    self.__didCallOnDeleteWidget = True
                    super(Cls, self).onDeleteWidget()
                __didCallOnDeleteWidget = False

            Cls.__name__ = cls.__name__
            Cls.__qualname__ = cls.__qualname__
            Cls.__module__ = cls.__module__

        if reset_default_settings:
            self.reset_default_settings(Cls)
        widget = Cls.__new__(Cls, signal_manager=self.signal_manager,
                             stored_settings=stored_settings)
        widget.__init__()
        self.process_events()
        self.widgets.append(widget)
        return widget

    @staticmethod
    def reset_default_settings(widget):
        """Reset default setting values for widget

        Discards settings read from disk and changes stored by fast_save

        Parameters
        ----------
        widget : OWBaseWidget
            widget to reset settings for
        """
        settings_handler = getattr(widget, "settingsHandler", None)
        if settings_handler:
            # Rebind settings handler to get fresh copies of settings
            # in known_settings
            settings_handler.bind(widget)
            # Reset defaults read from disk
            settings_handler.defaults = {}
            # Reset context settings
            settings_handler.global_contexts = []

    def process_events(self, until: callable = None, timeout=DEFAULT_TIMEOUT):
        """Process Qt events, optionally until `until` returns
        something True-ish.

        Needs to be called manually as QApplication.exec is never called.

        Parameters
        ----------
        until: callable or None
            If callable, the events are processed until the function returns
            something True-ish.
        timeout: int
            If until condition is not satisfied within timeout milliseconds,
            a TimeoutError is raised.

        Returns
        -------
        If until is not None, the True-ish result of its call.
        """
        if until is None:
            until = lambda: True

        started = time.perf_counter()
        while True:
            app.processEvents()
            try:
                result = until()
                if result:
                    return result
            except Exception:  # until can fail with anything; pylint: disable=broad-except
                pass
            if (time.perf_counter() - started) * 1000 > timeout:
                raise TimeoutError()
            time.sleep(.05)

    def show(self, widget=None):
        """Show widget in interactive mode.

        Useful for debugging tests, as widget can be inspected manually.
        """
        widget = widget or self.widget
        widget.show()
        app.exec()

    def send_signal(self, input, value=NO_VALUE, *args, widget=None, wait=-1):
        """ Send signal to widget by calling appropriate triggers.

        Parameters
        ----------
        input : str
        value : Object
        id : int
            channel id, used for inputs with flag Multiple
        widget : Optional[OWBaseWidget]
            widget to send signal to. If not set, self.widget is used
        wait : int
            The amount of time to wait for the widget to complete.
        """
        if value is NO_VALUE:
            value = input
            wid = widget or self.widget
            inputs = wid.inputs or wid.Inputs.__dict__.values()
            assert len(inputs) == 1
            input = next(iter(inputs))

        return self.send_signals([(input, value)], *args,
                                 widget=widget, wait=wait)

    def send_signals(self, signals, *args, widget=None, wait=-1):
        """ Send signals to widget by calling appropriate triggers.
        After all the signals are send, widget's handleNewSignals() in invoked.

        Parameters
        ----------
        signals : list of (str, Object)
        widget : Optional[OWBaseWidget]
            widget to send signals to. If not set, self.widget is used
        wait : int
            The amount of time to wait for the widget to complete.
        """
        if widget is None:
            widgets = {signal.widget for signal, _ in signals
                       if hasattr(signal, "widget")}
            if not widgets:
                widget = self.widget
            elif len(widgets) == 1:
                widget = widgets.pop()
            else:
                raise ValueError("Signals are bound to different widgets")

        for input, value in signals:
            self._send_signal(widget, input, value, *args)
        widget.handleNewSignals()
        if wait >= 0:
            self.wait_until_finished(widget, timeout=wait)

    @staticmethod
    def _send_signal(widget, input, value, *args, **kwargs):

        if isinstance(input, str):
            input = get_input_meta(widget, input)
            if input is None:
                raise ValueError("'{}' is not an input name for widget {}"
                                 .format(input, type(widget).__name__))
        if not widget.isReady():
            raise RuntimeError("'send_signal' called but the widget is not "
                               "in ready state and does not accept inputs.")
        # Assert sent input is of correct class
        assert isinstance(value, (input.type, type(None), type(input.closing_sentinel))), \
            '{} should be {}'.format(value.__class__.__mro__, input.type)
        notify_input_helper(input, widget, value, *args, **kwargs)

    def wait_until_stop_blocking(self, widget=None, wait=DEFAULT_TIMEOUT):
        """Wait until the widget stops blocking i.e. finishes computation.

        Parameters
        ----------
        widget : Optional[OWBaseWidget]
            widget to send signal to. If not set, self.widget is used
        wait : int
            The amount of time to wait for the widget to complete.

        """
        if widget is None:
            widget = self.widget

        if widget.isBlocking():
            spy = QSignalSpy(widget.blockingStateChanged)
            self.assertTrue(spy.wait(timeout=wait))

    def wait_until_finished(
            self, widget: Optional[OWBaseWidget] = None,
            timeout=DEFAULT_TIMEOUT) -> None:
        """Wait until the widget finishes computation.

        The widget is considered finished once all its outputs are valid.

        Parameters
        ----------
        widget : Optional[OWBaseWidget]
            widget to send signal to. If not set, self.widget is used
        timeout : int
            The amount of time to wait for the widget to complete.
        """
        if widget is None:
            widget = self.widget
        self.assertTrue(
            self.signal_manager.wait_for_finished(widget, timeout),
            f"Did not finish in the specified {timeout}ms timeout"
        )

    def commit_and_wait(self, widget=None, wait=DEFAULT_TIMEOUT):
        """Unconditional commit and wait until finished.

        Parameters
        ----------
        widget : Optional[OWBaseWidget]
            widget to send signal to. If not set, self.widget is used
        wait : int
            The amount of time to wait for the widget to complete.

        """
        if widget is None:
            widget = self.widget

        if hasattr(widget.commit, "now"):
            widget.commit.now()
        else:
            # Support for deprecated, non-decorated commit
            widget.unconditional_commit()
        self.wait_until_finished(widget=widget, timeout=wait)

    def get_output(self, output=None, widget=None, wait=DEFAULT_TIMEOUT):
        """Return the last output that has been sent from the widget.

        Parameters
        ----------
        output_name : str
        widget : Optional[OWBaseWidget]
            widget whose output is returned. If not set, self.widget is used
        wait : int
            The amount of time (in milliseconds) to wait for widget to complete.

        Returns
        -------
        The last sent value of given output or None if nothing has been sent.
        """
        if widget is None:
            # `output` may be an unbound signal with `widget` set to `None`
            # In this case, we use `self.widget`.
            widget = getattr(output, "widget", self.widget) or self.widget

        outputs = widget.outputs or widget.Outputs.__dict__.values()
        if output is None:
            assert len(outputs) == 1
            output = next(iter(outputs)).name
        elif not isinstance(output, str):
            output = output.name
        # widget.outputs are old-style signals; if empty, use new style
        assert output in (out.name for out in outputs), \
            "widget {} has no output {}".format(widget.name, output)
        value = widget.signalManager.get_output(widget, output, wait)
        if LazyValue.is_lazy(value):
            value = value.get_value()
        return value

    @contextmanager
    def modifiers(self, modifiers):
        """
        Context that simulates pressed modifiers

        Since QTest.keypress requries pressing some key, we simulate
        pressing "BassBoost" that looks exotic enough to not meddle with
        anything.
        """
        old_modifiers = QApplication.keyboardModifiers()
        try:
            QTest.keyPress(self.widget, Qt.Key_BassBoost, modifiers)
            yield
        finally:
            QTest.keyRelease(self.widget, Qt.Key_BassBoost, old_modifiers)

    def check_minimum_size(self, widget):

        def invalidate_cached_size_hint(w):
            # as in OWBaseWidget.setVisible
            if w.controlArea is not None:
                w.controlArea.updateGeometry()
            if w.buttonsArea is not None:
                w.buttonsArea.updateGeometry()
            if w.mainArea is not None:
                w.mainArea.updateGeometry()

        invalidate_cached_size_hint(widget)
        min_size = widget.minimumSizeHint()
        self.assertLess(min_size.width(), 800)
        self.assertLess(min_size.height(), 700)

    def check_msg_base_class(self, widget):
        """ Test whether widget error, warning and info messages are derived
        from its (direct) parent message classes. """
        def inspect(msg):
            msg_cls = getattr(widget, msg).__class__
            msg_base_cls = getattr(widget.__class__.__bases__[0], msg)
            self.assertTrue(issubclass(msg_cls, msg_base_cls))

        inspect("Error")
        inspect("Warning")
        inspect("Information")

    def check_export_image(self, widget):
        widget.copy_to_clipboard()

class TestWidgetTest(WidgetTest):
    """Meta tests for widget test helpers"""

    def test_process_events_handles_timeouts(self):
        with self.assertRaises(TimeoutError):
            self.process_events(until=lambda: False, timeout=0)

    def test_minimum_size(self):
        return  # skip this test

    def test_check_msg_base_class(self):
        class A(OWBaseWidget, openclass=True):
            pass

        class B(A):
            class Error(A.Error):
                pass

        class C(A, openclass=True):
            class Error(OWBaseWidget.Error):
                pass

        class D(C):
            class Error(A.Error):
                pass

        self.check_msg_base_class(B())
        self.check_msg_base_class(C())  # It is unfortunate that this passes...
        self.assertRaises(AssertionError, self.check_msg_base_class, D())

    def test_get_single_output(self):
        class A(OWBaseWidget):
            name = "A"

            class Inputs(OWBaseWidget.Inputs):
                question = Input("Question", str, auto_summary=False)

            class Outputs(OWBaseWidget.Outputs):
                answer = Output("Answer", int, auto_summary=False)

            @Inputs.question
            def question(self, s):
                self.Outputs.answer.send(eval(s))

        self.widget = self.create_widget(A)

        self.assertIsNone(self.get_output())
        self.send_signal("6 * 7")
        self.assertEquals(self.get_output(), 42)

    def test_compute_lazy_signals(self):
        class A(OWBaseWidget):
            name = "A"

            class Outputs(OWBaseWidget.Outputs):
                answer = Output("Answer", int, auto_summary=False)

            def __init__(self):
                self.Outputs.answer.send(LazyValue[int](lambda: 42))

        self.widget = self.create_widget(A)
        self.assertEquals(self.get_output(), 42)


class BaseParameterMapping:
    """Base class for mapping between gui components and learner's parameters
    when testing learner widgets.

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    gui_element : QWidget
        Gui component who's corresponding parameter is to be tested.

    values: list
        List of values to be tested.

    getter: function
        It gets component's value.

    setter: function
        It sets component's value.
    """

    def __init__(self, name, gui_element, values, getter, setter,
                 problem_type="both"):
        self.name = name
        self.gui_element = gui_element
        self.values = values
        self.get_value = getter
        self.set_value = setter
        self.problem_type = problem_type

    def __str__(self):
        if self.problem_type == "both":
            return self.name
        else:
            return "%s (%s)" % (self.name, self.problem_type)


class DefaultParameterMapping(BaseParameterMapping):
    """Class for mapping between gui components and learner's parameters
    when testing unchecked properties and therefore default parameters
    should be used.

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    default_value: str, int,
        Value that should be used by default.
    """

    def __init__(self, name, default_value):
        super().__init__(name, None, [default_value],
                         lambda: default_value, lambda x: None)


class ParameterMapping(BaseParameterMapping):
    """Class for mapping between gui components and learner parameters
    when testing learner widgets

    Parameters
    ----------
    name : str
        Name of learner's parameter.

    gui_element : QWidget
        Gui component who's corresponding parameter is to be tested.

    values: list, mandatory for ComboBox, optional otherwise
        List of values to be tested. When None, it is set according to
        component's type.

    getter: function, optional
        It gets component's value. When None, it is set according to
        component's type.

    setter: function, optional
        It sets component's value. When None, it is set according to
        component's type.
    """

    def __init__(self, name, gui_element, values=None,
                 getter=None, setter=None, **kwargs):
        super().__init__(
            name, gui_element,
            values or self._default_values(gui_element),
            getter or self._default_get_value(gui_element, values),
            setter or self._default_set_value(gui_element, values),
            **kwargs)

    @staticmethod
    def get_gui_element(widget, attribute):
        return widget.controlled_attributes[attribute][0].control

    @classmethod
    def from_attribute(cls, widget, attribute, parameter=None):
        return cls(parameter or attribute, cls.get_gui_element(widget, attribute))

    @staticmethod
    def _default_values(gui_element):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return [gui_element.minimum(), gui_element.maximum()]
        else:
            raise TypeError("{} is not supported".format(gui_element))

    @staticmethod
    def _default_get_value(gui_element, values):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return lambda: gui_element.value()
        elif isinstance(gui_element, QComboBox):
            return lambda: values[gui_element.currentIndex()]
        else:
            raise TypeError("{} is not supported".format(gui_element))

    @staticmethod
    def _default_set_value(gui_element, values):
        if isinstance(gui_element, (QSpinBox, QDoubleSpinBox, QSlider)):
            return lambda val: gui_element.setValue(val)
        elif isinstance(gui_element, QComboBox):
            def fun(val):
                value = values.index(val)
                gui_element.activated.emit(value)
                gui_element.setCurrentIndex(value)

            return fun
        else:
            raise TypeError("{} is not supported".format(gui_element))


@contextmanager
def open_widget_classes():
    with patch.object(OWBaseWidget, "__init_subclass__"):
        yield


@contextmanager
def override_default_settings(widget, defaults=None, context_defaults=[], handler=None):
    if defaults is None:
        defaults = {}

    h = (handler or SettingsHandler)()
    h.widget_class = widget
    h.defaults = defaults
    filename = h._get_settings_filename()

    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "wb") as f:
        pickle.dump(defaults, f)
        pickle.dump(context_defaults, f)

    yield

    if os.path.isfile(filename):
        os.remove(filename)


if __name__ == "__main__":
    unittest.main()



from AnyQt.QtWidgets import (
    QPushButton, QAbstractButton, QStyle, QStylePainter, QStyleOptionButton,
    QStyleOption
)
from AnyQt.QtGui import QPalette, QIcon, QLinearGradient, QBrush, QPainter
from AnyQt.QtCore import Qt, QSize, QEvent


class VariableTextPushButton(QPushButton):
    """
    QPushButton subclass with an sizeHint method to better support settable
    variable width button text.

    Use this class instead of the QPushButton when the button will
    switch the text dynamically while displayed.
    """
    def __init__(self, *args, textChoiceList=[], **kwargs):
        super().__init__(*args, **kwargs)
        self.__textChoiceList = list(textChoiceList)

    def setTextChoiceList(self, textList):
        """
        Set the list of all `text` string to use for size hinting.

        Parameters
        ----------
        textList : List[str]
            A list of all different `text` properties that will/can be set on
            the push button. This list is used to derive a suitable sizeHint
            for the widget.
        """
        self.__textChoiceList = textList
        self.updateGeometry()

    def sizeHint(self):
        """
        Reimplemented from `QPushButton.sizeHint`.

        Returns
        -------
        sh : QSize
        """
        sh = super().sizeHint()
        option = QStyleOptionButton()
        self.initStyleOption(option)
        style = self.style()
        fm = option.fontMetrics
        if option.iconSize.isValid():
            icsize = option.iconSize
            icsize.setWidth(icsize.width() + 4)
        else:
            icsize = QSize()

        for text in self.__textChoiceList:
            option.text = text
            size = fm.size(Qt.TextShowMnemonic, text)

            if not icsize.isNull():
                size.setWidth(size.width() + icsize.width())
                size.setHeight(max(size.height(), icsize.height()))

            sh = sh.expandedTo(
                style.sizeFromContents(QStyle.CT_PushButton, option,
                                       size, self))
        return sh


class SimpleButton(QAbstractButton):
    """
    A simple icon button widget.
    """
    def event(self, event):
        if event.type() == QEvent.Enter or event.type() == QEvent.Leave:
            self.update()
        return super().event(event)

    def sizeHint(self):
        # reimplemented
        self.ensurePolished()
        iconsize = self.iconSize()
        icon = self.icon()
        if not icon.isNull():
            iconsize = icon.actualSize(iconsize)
        margins = self.contentsMargins()
        return QSize(
            iconsize.width() + margins.left() + margins.right(),
            iconsize.height() + margins.top() + margins.bottom()
        )

    def minimumSizeHint(self):
        # reimplemented
        return self.sizeHint()

    def initStyleOption(self, option: QStyleOptionButton):
        option.initFrom(self)
        option.features |= QStyleOptionButton.Flat
        option.iconSize = self.iconSize()
        option.icon = self.icon()
        option.text = ""
        if self.isDown():
            option.state |= QStyle.State_Sunken

    def paintEvent(self, event):
        painter = QStylePainter(self)
        option = QStyleOptionButton()
        self.initStyleOption(option)
        if self.isDown():
            # Use default pressed flat button look
            painter.drawPrimitive(QStyle.PE_PanelButtonBevel, option)
        elif option.state & (QStyle.State_MouseOver | QStyle.State_HasFocus):
            flat_button_hover_background(painter, option)

        if not option.icon.isNull():
            rect = option.rect.adjusted(1, 1, -1, -1)
            if option.state & QStyle.State_Active:
                mode = (QIcon.Active if option.state & QStyle.State_MouseOver
                        else QIcon.Normal)
            elif option.state & QStyle.State_Enabled:
                mode = QIcon.Normal
            else:
                mode = QIcon.Disabled
            if self.isChecked():
                state = QIcon.On
            else:
                state = QIcon.Off
            option.icon.paint(painter, rect, Qt.AlignCenter, mode, state)


def flat_button_hover_background(
        painter: QPainter, option: QStyleOption
):
    palette = option.palette
    g = QLinearGradient(0, 0, 0, 1)
    g.setCoordinateMode(QLinearGradient.ObjectBoundingMode)
    base = palette.color(QPalette.Window)
    base.setAlpha(200)
    if base.value() < 127:
        base_ = base.lighter(170)
    else:
        base_ = base.darker(130)
    g.setColorAt(0, base_)
    g.setColorAt(0.6, base)
    g.setColorAt(1.0, base_)
    brush = QBrush(base_)
    cg = palette_color_group(option.state)
    if option.state & QStyle.State_HasFocus:
        pen = palette.color(cg, QPalette.Highlight)
    elif option.state & QStyle.State_MouseOver:
        pen = palette.color(cg, QPalette.WindowText)
        pen.setAlpha(50)
    else:
        pen = Qt.NoPen
    painter.save()
    painter.setPen(pen)
    painter.setBrush(brush)
    painter.translate(0.5, 0.5)
    painter.setRenderHints(QPainter.Antialiasing, True)
    painter.drawRoundedRect(option.rect.adjusted(0, 0, -1, -1), 2., 2., )
    painter.restore()


def palette_color_group(state: QStyle.StateFlag) -> QPalette.ColorGroup:
    if not state & QStyle.State_Enabled:
        return QPalette.Disabled
    elif state & QStyle.State_Active:
        return QPalette.Active
    else:
        return QPalette.Inactive



# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE SETUP.PY
short_version = '4.22.0'
version = '4.22.0'
full_version = '4.22.0'
git_revision = 'afa06382ea8e6635cad86d3f6ead16bdc8548963'
release = True

if not release:
    version = full_version
    short_version += ".dev"



import copy
import itertools
import warnings
from collections import defaultdict
from functools import singledispatch
import inspect
from typing import (
    NamedTuple, Union, Optional, Iterable, Dict, Tuple, Any, Sequence,
    Callable
)

from AnyQt.QtCore import Qt, pyqtSignal, QPoint
from AnyQt.QtWidgets import QWidgetAction, QMenu, QWidget, QLabel, QVBoxLayout

from orangecanvas.registry.description import (
    InputSignal, OutputSignal, Single, Multiple, Default, NonDefault,
    Explicit, Dynamic
)
# imported here for easier use by widgets, pylint: disable=unused-import
from orangecanvas.scheme.signalmanager import LazyValue

from orangewidget.utils.messagewidget import MessagesWidget
from orangewidget.workflow.utils import WeakKeyDefaultDict


# increasing counter for ensuring the order of Input/Output definitions
# is preserved when going through the unordered class namespace of
# WidgetSignalsMixin.Inputs/Outputs.
_counter = itertools.count()


class PartialSummary(NamedTuple):
    summary: Union[None, str, int] = None
    details: Optional[str] = None
    preview_func: Optional[Callable[[], QWidget]] = None


def base_summarize(_) -> PartialSummary:
    return PartialSummary()


summarize = singledispatch(base_summarize)
summarize.__doc__ = """
Function for summarizing the input or output data.

The function must be decorated with `@summarize.register`. It accepts an
argument of arbitrary type and returns a `PartialSummary`, which is a tuple
consisting of two strings: a short summary (usually a number) and details.
"""


SUMMARY_STYLE = """
<style>
    ul {
        margin-left: 4px;
        margin-top: 2px;
        -qt-list-indent:1
    }

    li {
        margin-bottom: 3px;
    }

    th {
        text-align: right;
    }
</style>
"""


def can_summarize(type_, name, explicit):
    if explicit is not None:
        return explicit
    if not isinstance(type_, tuple):
        type_ = (type_, )
    instr = f"To silence this warning, set auto_summary of '{name}' to False."
    for a_type in type_:
        if isinstance(a_type, str):
            warnings.warn(
                f"Output is specified with qualified name ({a_type}). "
                "To enable auto summary, set auto_summary to True. "
                + instr, UserWarning)
            return False
        if summarize.dispatch(a_type) is base_summarize:
            warnings.warn(
                f"register 'summarize' function for type {a_type.__name__}. "
                + instr, UserWarning, stacklevel=4)
            return False
    return True


Closed = type(
    "Closed", (object,), {
        "__doc__": "Explicit connection closing sentinel.",
        "__repr__": lambda self: "Closed",
        "__str__": lambda self: "Closed",
    }
)()


class _Signal:
    @staticmethod
    def get_flags(multiple, default, explicit, dynamic):
        """Compute flags from arguments"""

        return (Multiple if multiple else Single) | \
                (Default if default else NonDefault) | \
                (explicit and Explicit) | \
                (dynamic and Dynamic)

    def bound_signal(self, widget):
        """
        Return a copy of the signal bound to a widget.

        Called from `WidgetSignalsMixin.__init__`
        """
        new_signal = copy.copy(self)
        new_signal.widget = widget
        return new_signal


def getsignals(signals_cls):
    # This function is preferred over getmembers because it returns the signals
    # in order of appearance
    return [(k, v)
            for cls in reversed(inspect.getmro(signals_cls))
            for k, v in cls.__dict__.items()
            if isinstance(v, _Signal)]


class Input(InputSignal, _Signal):
    """
    Description of an input signal.

    The class is used to declare input signals for a widget as follows
    (the example is taken from the widget Test & Score)::

        class Inputs:
            train_data = Input("Data", Table, default=True)
            test_data = Input("Test Data", Table)
            learner = Input("Learner", Learner, multiple=True)
            preprocessor = Input("Preprocessor", Preprocess)

    Every input signal must be used to decorate exactly one method that
    serves as the input handler, for instance::

        @Inputs.train_data
        def set_train_data(self, data):
            ...

    Parameters
    ----------
    name (str):
        signal name
    type (type):
        signal type
    id (str):
        a unique id of the signal
    doc (str, optional):
        signal documentation
    replaces (list of str):
        a list with names of signals replaced by this signal
    multiple (bool, optional):
        if set, multiple signals can be connected to this output
        (default: `False`)
    default (bool, optional):
        when the widget accepts multiple signals of the same type, one of them
        can set this flag to act as the default (default: `False`)
    explicit (bool, optional):
        if set, this signal is only used when it is the only option or when
        explicitly connected in the dialog (default: `False`)
    auto_summary (bool, optional):
        by default, the input is reflected in widget's summary for all types
        with registered `summarize` function. This can be overridden by
        explicitly setting `auto_summary` to `False` or `True`. Explicitly
        setting this argument will also silence warnings for types without
        the summary function and for types defined with a fully qualified
        string instead of an actual type object.
    """
    Closed = Closed

    def __init__(self, name, type, id=None, doc=None, replaces=None, *,
                 multiple=False, default=False, explicit=False,
                 auto_summary=None, closing_sentinel=None):
        flags = self.get_flags(multiple, default, explicit, False)
        super().__init__(name, type, "", flags, id, doc, replaces or [])
        self.auto_summary = can_summarize(type, name, auto_summary)
        self._seq_id = next(_counter)
        self.closing_sentinel = closing_sentinel

    def __call__(self, method):
        """
        Decorator that stores decorated method's name in the signal's
        `handler` attribute. The method is returned unchanged.
        """
        if self.flags & Multiple:
            def summarize_wrapper(widget, value, id=None):
                # If this method is overridden, don't summarize
                if summarize_wrapper is getattr(type(widget), method.__name__):
                    widget.set_partial_input_summary(
                        self.name, summarize(value), id=id)
                method(widget, value, id)
        else:
            def summarize_wrapper(widget, value):
                if summarize_wrapper is getattr(type(widget), method.__name__):
                    widget.set_partial_input_summary(
                        self.name, summarize(value))
                method(widget, value)

        # Re-binding with the same name can happen in derived classes
        # We do not allow re-binding to a different name; for the same class
        # it wouldn't work, in derived class it could mislead into thinking
        # that the signal is passed to two different methods
        if self.handler and self.handler != method.__name__:
            raise ValueError("Input {} is already bound to method {}".
                             format(self.name, self.handler))
        self.handler = method.__name__
        return summarize_wrapper if self.auto_summary else method


class MultiInput(Input):
    """
    A special multiple input descriptor.

    This type of input has explicit set/insert/remove interface to maintain
    fully ordered sequence input. This should be preferred to the
    plain `Input(..., multiple=True)` descriptor.

    This input type must register three methods in the widget implementation
    class corresponding to the insert, set/update and remove input commands::

        class Inputs:
            values = MultiInput("Values", object)
        ...
        @Inputs.values
        def set_value(self, index: int, value: object):
            "Set/update the value at index"
            ...
        @Inputs.values.insert
        def insert_value(self, index: int, value: object):
            "Insert value at specified index"
            ...
        @Inputs.values.remove
        def remove_value(self, index: int):
            "Remove value at index"
            ...

    Parameters
    ----------
    filter_none: bool
        If `True` any `None` values sent by workflow execution
        are implicitly converted to 'remove' notifications. When the value
        again changes to non-None the input is re-inserted into its proper
        position.


    .. versionadded:: 4.13.0
    """
    insert_handler: str = None
    remove_handler: str = None

    def __init__(self, *args, filter_none=False, **kwargs):
        multiple = kwargs.pop("multiple", True)
        if not multiple:
            raise ValueError("multiple cannot be set to False")
        super().__init__(*args, multiple=True, **kwargs)
        self.filter_none = filter_none
        self.closing_sentinel = Closed

    __summary_ids_mapping = WeakKeyDefaultDict(dict)
    __id_gen = itertools.count()

    def __get_summary_ids(self, widget: 'WidgetSignalsMixin'):
        ids = self.__summary_ids_mapping[widget]
        return ids.setdefault(self.name, [])

    def __call__(self, method):
        def summarize_wrapper(widget, index, value):
            # If this method is overridden, don't summarize
            if summarize_wrapper is getattr(type(widget), method.__name__):
                ids = self.__get_summary_ids(widget)
                widget.set_partial_input_summary(
                    self.name, summarize(value), id=ids[index], index=index)
            method(widget, index, value)
        _ = super().__call__(method)
        return summarize_wrapper if self.auto_summary else method

    def insert(self, method):
        """Register the method as the insert handler"""
        def summarize_wrapper(widget, index, value):
            if summarize_wrapper is getattr(type(widget), method.__name__):
                ids = self.__get_summary_ids(widget)
                ids.insert(index, next(self.__id_gen))
                widget.set_partial_input_summary(
                    self.name, summarize(value), id=ids[index], index=index)
            method(widget, index, value)
        self.insert_handler = method.__name__
        return summarize_wrapper if self.auto_summary else method

    def remove(self, method):
        """"Register the method as the remove handler"""
        def summarize_wrapper(widget, index):
            if summarize_wrapper is getattr(type(widget), method.__name__):
                ids = self.__get_summary_ids(widget)
                id_ = ids.pop(index)
                widget.set_partial_input_summary(
                    self.name, summarize(None), id=id_)
            method(widget, index)
        self.remove_handler = method.__name__
        return summarize_wrapper if self.auto_summary else method

    def bound_signal(self, widget):
        if self.insert_handler is None:
            raise RuntimeError('insert_handler is not set')
        if self.remove_handler is None:
            raise RuntimeError('remove_handler is not set')
        return super().bound_signal(widget)


_not_set = object()


def _parse_call_id_arg(id=_not_set):
    if id is _not_set:
        return None
    else:
        warnings.warn(
            "`id` parameter is deprecated and will be removed in the "
            "future", FutureWarning, stacklevel=3,
        )
        return id


class Output(OutputSignal, _Signal):
    """
    Description of an output signal.

    The class is used to declare output signals for a widget as follows
    (the example is taken from the widget Test & Score)::

        class Outputs:
            predictions = Output("Predictions", Table)
            evaluations_results = Output("Evaluation Results", Results)

    The signal is then transmitted by, for instance::

        self.Outputs.predictions.send(predictions)

    Parameters
    ----------
    name (str):
        signal name
    type (type):
        signal type
    id (str):
        a unique id of the signal
    doc (str, optional):
        signal documentation
    replaces (list of str):
        a list with names of signals replaced by this signal
    default (bool, optional):
        when the widget accepts multiple signals of the same type, one of them
        can set this flag to act as the default (default: `False`)
    explicit (bool, optional):
        if set, this signal is only used when it is the only option or when
        explicitly connected in the dialog (default: `False`)
    dynamic (bool, optional):
        Specifies that the instances on the output will in general be subtypes
        of the declared type and that the output can be connected to any input
        signal which can accept a subtype of the declared output type
        (default: `True`)
    auto_summary (bool, optional):
        by default, the output is reflected in widget's summary for all types
        with registered `summarize` function. This can be overridden by
        explicitly setting `auto_summary` to `False` or `True`. Explicitly
        setting this argument will also silence warnings for types without
        the summary function and for types defined with a fully qualified
        string instead of an actual type object.
    """
    def __init__(self, name, type, id=None, doc=None, replaces=None, *,
                 default=False, explicit=False, dynamic=True,
                 auto_summary=None):
        flags = self.get_flags(False, default, explicit, dynamic)
        super().__init__(name, type, flags, id, doc, replaces or [])
        self.auto_summary = can_summarize(type, name, auto_summary)
        self.widget = None
        self._seq_id = next(_counter)

    def send(self, value, *args, **kwargs):
        """Emit the signal through signal manager."""
        assert self.widget is not None
        id = _parse_call_id_arg(*args, **kwargs)
        signal_manager = self.widget.signalManager
        if signal_manager is not None:
            if id is not None:
                extra_args = (id,)
            else:
                extra_args = ()
            signal_manager.send(self.widget, self.name, value, *extra_args)
        if self.auto_summary:
            self.widget.set_partial_output_summary(
                self.name, summarize(value), id=id)

    def invalidate(self):
        """Invalidate the current output value on the signal"""
        assert self.widget is not None
        signal_manager = self.widget.signalManager
        if signal_manager is not None:
            signal_manager.invalidate(self.widget, self.name)


class WidgetSignalsMixin:
    """Mixin for managing widget's input and output signals"""
    class Inputs:
        pass

    class Outputs:
        pass

    def __init__(self):
        self.input_summaries = {}
        self.output_summaries: Dict[str, PartialSummary] = {}
        self._bind_signals()

    def _bind_signals(self):
        for direction, summaries in (("Inputs", self.input_summaries),
                                     ("Outputs", self.output_summaries)):
            bound_cls = getattr(self, direction)
            bound_signals = bound_cls()
            for name, signal in getsignals(bound_cls):
                setattr(bound_signals, name, signal.bound_signal(self))
                if signal.auto_summary:
                    summaries[signal.name] = {}
            setattr(self, direction, bound_signals)

    def send(self, signalName, value, *args, **kwargs):
        """
        Send a `value` on the `signalName` widget output.

        An output with `signalName` must be defined in the class ``outputs``
        list.
        """
        id = _parse_call_id_arg(*args, **kwargs)
        if not any(s.name == signalName for s in self.outputs):
            raise ValueError('{} is not a valid output signal for widget {}'.format(
                signalName, self.name))

        if self.signalManager is not None:
            if id is not None:
                extra_args = (id,)
            else:
                extra_args = ()
            self.signalManager.send(self, signalName, value, *extra_args)

    def handleNewSignals(self):
        """
        Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        Reimplement this method in order to coalesce updates from
        multiple updated inputs.
        """
        pass

    # Methods used by the meta class
    @classmethod
    def convert_signals(cls):
        """
        Maintenance and sanity checks for signals.

        - Convert tuple descriptions into old-style signals for backwards compatibility
        - For new-style in classes, copy attribute name to id, if id is not set explicitly
        - Check that all input signals have handlers
        - Check that the same name and/or does not refer to different signals.

        This method is called from the meta-class.
        """
        def signal_from_args(args, signal_type):
            if isinstance(args, tuple):
                return signal_type(*args)
            elif isinstance(args, signal_type):
                return copy.copy(args)

        if hasattr(cls, "inputs") and cls.inputs:
            cls.inputs = [signal_from_args(input_, InputSignal)
                          for input_ in cls.inputs]
        if hasattr(cls, "outputs") and cls.outputs:
            cls.outputs = [signal_from_args(output, OutputSignal)
                           for output in cls.outputs]

        for direction in ("Inputs", "Outputs"):
            klass = getattr(cls, direction, None)
            if klass is None:
                continue
            for name, signal in klass.__dict__.items():
                if isinstance(signal, (_Signal)) and signal.id is None:
                    signal.id = name

        cls._check_input_handlers()
        cls._check_ids_unique()

    @classmethod
    def _check_input_handlers(cls):
        unbound = [signal.name
                   for _, signal in getsignals(cls.Inputs)
                   if not signal.handler]
        if unbound:
            raise ValueError("unbound signal(s) in {}: {}".
                             format(cls.__name__, ", ".join(unbound)))

        missing_handlers = [signal.handler for signal in cls.inputs
                            if not hasattr(cls, signal.handler)]
        if missing_handlers:
            raise ValueError("missing handlers in {}: {}".
                             format(cls.__name__, ", ".join(missing_handlers)))

    @classmethod
    def _check_ids_unique(cls):
        for direction in ("input", "output"):
            # Collect signals by name and by id, check for duplicates
            by_name = {}
            by_id = {}
            for signal in cls.get_signals(direction + "s"):
                if signal.name in by_name:
                    raise RuntimeError(
                        f"Name {signal.name} refers to different {direction} "
                        f"signals of {cls.__name__}" )
                by_name[signal.name] = signal
                if signal.id is not None:
                    if signal.id in by_id:
                        raise RuntimeError(
                            f"Id {signal.id} refers to different {direction} "
                            f"signals of {cls.__name__}" )
                    by_id[signal.id] = signal

            # Warn the same name and id refer to different signal
            for name in set(by_name) & set(by_id):
                if by_name[name] is not by_id[name]:
                    warnings.warn(
                        f"{name} appears as a name and an id of two different "
                        f"{direction} signals in {cls.__name__}")

    @classmethod
    def get_signals(cls, direction, ignore_old_style=False):
        """
        Return a list of `InputSignal` or `OutputSignal` needed for the
        widget description. For old-style signals, the method returns the
        original list. New-style signals are collected into a list.

        Parameters
        ----------
        direction (str): `"inputs"` or `"outputs"`

        Returns
        -------
        list of `InputSignal` or `OutputSignal`
        """
        old_style = cls.__dict__.get(direction, None)
        if old_style and not ignore_old_style:
            return old_style

        signal_class = getattr(cls, direction.title())
        signals = [signal for _, signal in getsignals(signal_class)]
        return list(sorted(signals, key=lambda s: s._seq_id))

    def update_summaries(self):
        self._update_summary(self.input_summaries)
        self._update_summary(self.output_summaries)

    def set_partial_input_summary(self, name, partial_summary, *,
                                  id=None, index=None):
        self.__set_part_summary(self.input_summaries[name], id, partial_summary, index=index)
        self._update_summary(self.input_summaries)

    def set_partial_output_summary(self, name, partial_summary, *, id=None):
        self.__set_part_summary(self.output_summaries[name], id, partial_summary)
        self._update_summary(self.output_summaries)

    @staticmethod
    def __set_part_summary(summary, id, partial_summary, index=None):
        if partial_summary.summary is None:
            if id in summary:
                del summary[id]
        else:
            if index is None or id in summary:
                summary[id] = partial_summary
            else:
                # Insert inplace at specified index
                items = list(summary.items())
                items.insert(index, (id, partial_summary))
                summary.clear()
                summary.update(items)

    def _update_summary(self, summaries):
        from orangewidget.widget import StateInfo

        def format_short(partial):
            summary = partial.summary
            if summary is None:
                return "-"
            if isinstance(summary, int):
                return StateInfo.format_number(summary)
            if isinstance(summary, str):
                return summary
            raise ValueError("summary must be None, string or int; "
                             f"got {type(summary).__name__}")

        def format_detail(partial):
            if partial.summary is None:
                return "-"
            return str(partial.details or partial.summary)

        def join_multiples(partials):
            if not partials:
                return "-", "-"
            shorts = " ".join(map(format_short, partials.values()))
            details = "<br/>".join(format_detail(partial) for partial in partials.values())
            return shorts, details

        info = self.info
        is_input = summaries is self.input_summaries
        assert is_input or summaries is self.output_summaries

        if not summaries:
            return
        if not any(summaries.values()):
            summary = info.NoInput if is_input else info.NoOutput
            detail = ""
        else:
            summary, details = zip(*map(join_multiples, summaries.values()))
            summary = " | ".join(summary)
            detail = "<hr/><table>" \
                     + "".join(f"<tr><th><nobr>{name}</nobr>: "
                               f"</th><td>{detail}</td></tr>"
                               for name, detail in zip(summaries, details)) \
                     + "</table>"

        setter = info.set_input_summary if is_input else info.set_output_summary
        if detail:
            setter(summary, SUMMARY_STYLE + detail, format=Qt.RichText)
        else:
            setter(summary)

    def show_preview(self, summaries):
        view = QWidget(self)
        view.setLayout(QVBoxLayout())

        for name, summary in summaries.items():
            if not summary:
                view.layout().addWidget(QLabel("<hr/><table>"
                           f"<tr><th><nobr>{name}</nobr>: "
                           f"</th><td>-</td></tr>"
                           "</table>"))
            for i, part in enumerate(summary.values(), start=1):
                part_no = f" ({i})" if len(summary) > 1 else ""
                detail = str(part.details or part.summary) or "-"
                view.layout().addWidget(
                    QLabel("<hr/><table>"
                           f"<tr><th><nobr>{name}{part_no}</nobr>: "
                           f"</th><td>{detail}</td></tr>"
                           "</table>")
                )
                if part.preview_func:
                    preview = part.preview_func()
                    view.layout().addWidget(preview, 1)
        if view.layout().isEmpty():
            return
        view.layout().addStretch()

        screen = self.windowHandle().screen()
        geometry = screen.availableGeometry()
        preview = QMenu(self)
        wid, hei = geometry.width(), geometry.height()
        preview.setFixedSize(wid // 2, hei // 2)
        view.setFixedSize(wid // 2 - 4, hei // 2 - 4)
        action = QWidgetAction(preview)
        action.setDefaultWidget(view)
        preview.addAction(action)
        preview.popup(QPoint(wid // 4, hei // 4), action)


def get_input_meta(widget: WidgetSignalsMixin, name: str) -> Optional[Input]:
    """
    Return the named input meta description from widget (if it exists).
    """
    def as_input(obj):
        if isinstance(obj, Input):
            return obj
        elif isinstance(obj, InputSignal):
            rval = Input(obj.name, obj.type, obj.id, obj.doc, obj.replaces,
                         multiple=not obj.single, default=obj.default,
                         explicit=obj.explicit)
            rval.handler = obj.handler
            return rval
        elif isinstance(obj, tuple):
            return as_input(InputSignal(*obj))
        else:
            raise TypeError

    inputs: Iterable[Input] = map(as_input, widget.get_signals("inputs"))
    for input_ in inputs:
        if input_.name == name:
            return input_
    return None


def get_widget_inputs(
        widget: WidgetSignalsMixin
) -> Dict[str, Sequence[Tuple[Any, Any]]]:
    state: Dict[str, Sequence[Tuple[Any, Any]]]
    state = widget.__dict__.setdefault(
        "_WidgetSignalsMixin__input_state", {}
    )
    return state


@singledispatch
def notify_input_helper(
        input: Input, widget: WidgetSignalsMixin, obj, key=None, index=-1
) -> None:
    """
    Set the input to the `widget` in a way appropriate for the `input` type.
    """
    raise NotImplementedError


@notify_input_helper.register(Input)
def set_input_helper(
        input: Input, widget: WidgetSignalsMixin, obj, key=None, index=-1
):
    handler = getattr(widget, input.handler)
    if input.single:
        args = (obj,)
    else:
        args = (obj, key)
    handler(*args)


@notify_input_helper.register(MultiInput)
def set_multi_input_helper(
        input: MultiInput, widget: WidgetSignalsMixin, obj, key=None, index=-1,
):
    """
    Set/update widget's input for a `MultiInput` input to obj.

    `key` must be a unique for an input slot to update.
    `index` defines the position where a new input (key that did not
    previously exist) is inserted. The default -1 indicates that the
    new input should be appended to the end. An input is removed by using
    inout.closing_sentinel as the obj.
    """
    inputs_ = get_widget_inputs(widget)
    inputs = inputs_.setdefault(input.name, ())
    filter_none = input.filter_none

    signal_old = None
    key_to_pos = {key: i for i, (key, _) in enumerate(inputs)}
    update = key in key_to_pos
    new = key not in key_to_pos
    remove = obj is input.closing_sentinel
    if new:
        if not 0 <= index < len(inputs):
            index = len(inputs)
    else:
        index = key_to_pos.get(key)
        assert index is not None

    inputs_updated = list(inputs)
    if new:
        inputs_updated.insert(index, (key, obj))
    elif remove:
        signal_old = inputs_updated.pop(index)
    else:
        signal_old = inputs_updated[index]
        inputs_updated[index] = (key, obj)

    inputs_[input.name] = tuple(inputs_updated)

    if filter_none:
        def filter_f(obj):
            return obj is None
    else:
        filter_f = None

    def local_index(
            key: Any, inputs: Sequence[Tuple[Any, Any]],
            filter: Optional[Callable[[Any], bool]] = None,
    ) -> Optional[int]:
        i = 0
        for k, obj in inputs:
            if key == k:
                return i
            elif filter is not None:
                i += int(not filter(obj))
            else:
                i += 1
        return None

    if filter_none:
        # normalize signal.value is None to Close signal.
        filtered = filter_f(obj)
        if new and filtered:
            # insert in inputs only (done above)
            return
        elif new:
            # Some inputs before this might be filtered invalidating the
            # effective index. Find appropriate index for insertion
            index = len([obj for _, obj in inputs[:index] if not filter_f(obj)])
        elif remove:
            if filter_f(signal_old[1]):
                # was already notified as removed, only remove from inputs (done above)
                return
            else:
                index = local_index(key, inputs, filter_f)
        elif update and filtered:
            if filter_f(signal_old[1]):
                # did not change; remains filtered
                return
            else:
                # remove it
                remove = True
                new = False
                index = local_index(key, inputs, filter_f)
                assert index is not None
        elif update:
            index = local_index(key, inputs, filter_f)

        if signal_old is not None and filter_f(signal_old[1]) and not filtered:
            # update with non-none value, substitute as new signal
            new = True
            remove = False
            index = local_index(key, inputs, filter_f)

    if new:
        handler = input.insert_handler
        args = (index, obj)
    elif remove:
        handler = input.remove_handler
        args = (index, )
    else:
        handler = input.handler
        args = (index, obj)
    assert index is not None
    handler = getattr(widget, handler)
    handler(*args)
